//-------------------
// RB Plugin Example
//-------------------

//includes

#include <math.h>
#include <ctype.h>
#include <limits.h>
#include <setjmp.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rb_plugin.h"

#include "pcre.c"

//-----------------------
// PCRE functions
//-----------------------

int PCRECompileOptions(Boolean anchored, Boolean caseless, Boolean utf8) {
	int options = 0;
	if (anchored) options |= PCRE_ANCHORED;
	if (caseless) options |= PCRE_CASELESS;
	if (utf8) options |= PCRE_UTF8;
	return options;
}

int PCRECompile(REALstring pattern, int options) {
	const char* errptr;
	int erroffset;
	const char* ptn = REALCString(pattern);
	pcre* pcre = pcre_compile(ptn, options, &errptr, &erroffset, (const unsigned char*)NULL);
	return (int)pcre;
}

void PCREFree(void** ptr) {
	free(*ptr);
	*ptr = NULL;
}

REALstring PCREGetToken(const int pattern, REALstring subject, int start, int* strstart, int* strend) {
	//get substring count
	int ssCnt;
	pcre_fullinfo((const pcre*)pattern, (const pcre_extra*)NULL, PCRE_INFO_CAPTURECOUNT, &ssCnt);
	ssCnt *= 3;
	//create destination vector
	int vector[30];
	//subject string
	char* sbj = (char*)REALCString(subject);
	int subjlen = (int)strlen(sbj);
	//options
	int options = PCRE_ANCHORED | PCRE_NOTEMPTY;
	//call function
	int count = pcre_exec((const pcre*)pattern, (const pcre_extra*)NULL, sbj, subjlen, start, options, vector, ssCnt);
	//get string
	if (count == -1) return REALBuildString((char*)NULL, 0);
	//int st = vector[0];
	//int ln = vector[1] - st;
	char str[] = "a";
	//return realstring
	char* substring_start = sbj + vector[0];
	int substring_length = vector[1] - vector[0];
	*strstart = vector[0];
	*strend = vector[1];
	return REALBuildString(str, 2);
	return REALBuildString(substring_start, substring_length);
}

int PCREExec(int pattern, int extra, REALmemoryBlock subject, int start, int options, REALmemoryBlock vector) {
	char* subj = (char*)REALMemoryBlockGetPtr(subject);
	int subjsz = REALMemoryBlockGetSize(subject);
	int* vect = (int*)REALMemoryBlockGetPtr(vector);
	int vecsz = REALMemoryBlockGetSize(vector)/4;
	return pcre_exec((const pcre*)pattern, (const pcre_extra*)extra, subj, subjsz, start, options, vect, vecsz);
}

//-----------------------
// Utility functions
//-----------------------

static short ShortSwap(short s)
{
  unsigned char b1, b2;
  
  b1 = s & 255;
  b2 = (s >> 8) & 255;

  return (b1 << 8) + b2;
}

static int LongSwap (int i)
{
  unsigned char b1, b2, b3, b4;

  b1 = i & 255;
  b2 = ( i >> 8 ) & 255;
  b3 = ( i>>16 ) & 255;
  b4 = ( i>>24 ) & 255;

  return ((int)b1 << 24) + ((int)b2 << 16) + ((int)b3 << 8) + b4;
}

static float FloatSwap(float f)
{
  union
  {
    float f;
    unsigned char b[4];
  } dat1, dat2;

  dat1.f = f;
  dat2.b[0] = dat1.b[3];
  dat2.b[1] = dat1.b[2];
  dat2.b[2] = dat1.b[1];
  dat2.b[3] = dat1.b[0];
  return dat2.f;
}

static double DoubleSwap(double d)
{
  union
  {
    double d;
    unsigned char b[8];
  } dat1, dat2;

  dat1.d = d;
  dat2.b[0] = dat1.b[7];
  dat2.b[1] = dat1.b[6];
  dat2.b[2] = dat1.b[5];
  dat2.b[3] = dat1.b[4];
  dat2.b[4] = dat1.b[3];
  dat2.b[5] = dat1.b[2];
  dat2.b[6] = dat1.b[1];
  dat2.b[7] = dat1.b[0];
  return dat2.d;
}

static void BlockCopy(void *dst, void *src, long n) {
	#if TARGET_MACOS
		BlockMove(src, dst, n);
	#else
		long m = n/8;
		while (m--) *((double *)dst)++ = *((double *)src)++; 
		m = n%8;
		while (m--) *((char *)dst)++ = *((char *)src)++;
	#endif
}

static int Compare(void* ptr1, void* ptr2, long n) {
	long m = n/4;
	while (m--) if (*((long *)ptr1)++ != *((long *)ptr2)++) return false; 
	m = n%4;
	while (m--) if (*((char *)ptr1)++ != *((char *)ptr2)++) return false;
	return true;
}

//-----------------------
// Add global RB methods
//-----------------------

static REALmethodDefinition pluginMethodArray[] = {
	{ (REALproc) PCREFree, REALnoImplementation, "PCREFree(byref pointer as Integer)" },
	{ (REALproc) PCREExec, REALnoImplementation, "PCREExec(pattern as Integer, extra as Integer, subject as Memoryblock, start as Integer, options as Integer, vector as Memoryblock) as Integer" },
	{ (REALproc) PCRECompileOptions, REALnoImplementation, "PCRECompileOptions(anchored as Boolean, caseless as Boolean, utf8 as Boolean) as integer" },
	{ (REALproc) PCRECompile, REALnoImplementation, "PCRECompile(pattern as String, options as Integer) as integer" },
	{ (REALproc) PCREGetToken, REALnoImplementation, "PCREGetToken(pattern as Integer, subject as String, start as Integer, byref strstart as Integer, byref strend as Integer) as String" },
}; 

//------------------------------------------------------
// Create a new RB class or extend an existing RB class
//------------------------------------------------------

	// If you want to create a new class, set the following macro
	// to false;
	// If you want to extend an existing RB class, such as FolderItem,
	// Date or Application, set the value to true:

#define classExtension false

	// define storage space for the class

extern REALclassDefinition TokenTypeClass;

struct TokenTypeClassData
{
	int			id;
	REALstring  name;
	REALstring	pattern;
	REALstring  replacement;
};

	//class constructor/destructor

static void TokenTypeClassConstructor (REALobject instance)
{
	ClassData (TokenTypeClass, instance, TokenTypeClassData, me);

	// you can initialize objects of "me" here ...
	
	//me->integerProperty = 1234;
	//me->booleanProperty = true;
}

static void TokenTypeClassDestructor (REALobject instance)
{
	ClassData (TokenTypeClass, instance, TokenTypeClassData, me);

	// close & dispose objects of "me" here
	// (e.g. if you opened a file or allocated memory, you need to release it here again)
}

	//example custom getter/setter. used with callbacks to trigger an event when a property's value is changed

/*

static int ExampleIntGetter (REALobject instance, long param)
// This is called when RB reads a property of your class
{
	ClassData (ExampleClass, instance, ExampleClassData, me);
	return me->integerProperty;
}

static void ExampleIntSetter (REALobject instance, long param, int value)
// This is called when RB assigns a value to a property of your class
{
	ClassData (ExampleClass, instance, ExampleClassData, me);
	me->integerProperty = callExampleEvent (instance, value); //<- note here that the callback routine is called
}

	//example methods
	
static void ExampleClassMethod1 (REALobject instance, int value)
// This is called when RB calls a method of your class
{
	ClassData (ExampleClass, instance, ExampleClassData, me);
	me->integerProperty = value;
}

static void ExampleClassMethod2 (REALobject instance, int a, double b)
// This is called when RB calls a method of your class
{
	ClassData (ExampleClass, instance, ExampleClassData, me);
	me->doubleProperty = a*b;
}

	//register the class, its properties and its methods

static REALproperty ExampleClassProperties[] = {
	{ nil, "integerProperty", "Integer", 0, (REALproc) ExampleIntGetter, (REALproc) ExampleIntSetter, 0 },
	{ nil, "booleanProperty", "Boolean", 0, REALstandardGetter, REALstandardSetter, FieldOffset(ExampleClassData, booleanProperty) },
	{ nil, "doubleProperty", "Double", 0, REALstandardGetter, REALstandardSetter, FieldOffset(ExampleClassData, doubleProperty) },
};

static REALmethodDefinition ExampleClassMethods[] = {
	{ (REALproc) ExampleClassMethod1, REALnoImplementation, "ExampleClassMethod1(value as Integer)"},
	{ (REALproc) ExampleClassMethod2, REALnoImplementation, "ExampleClassMethod2(a as Integer, b as Double)"},
};

static REALclassDefinition ExampleClass = {
	kCurrentREALControlVersion,
	#if classExtension
		"aRBClassName",	// <- put the name of the to be extended class here, like "FolderItem"
		nil,
	#else
		"ExampleClass",	// <- this is the name of the class you're adding with this plugin
		nil,			// <- name of its super class, in case this is not a base class
	#endif
	sizeof(ExampleClassData),
	0,
	#if classExtension
		// Unfortunately, for extended classes, we don't get called when an obj is con-/destructed.
		// (this is a current RB restriction and might be changed in a later version, hopefully)
		nil, nil,
	#else
		(REALproc) ExampleClassConstructor, (REALproc) ExampleClassDestructor,
	#endif
	ExampleClassProperties,
	sizeof(ExampleClassProperties) / sizeof(REALproperty),
	ExampleClassMethods,
	sizeof(ExampleClassMethods) / sizeof(REALmethodDefinition),
	ExampleClassEvents,
	sizeof(ExampleClassEvents) / sizeof(REALevent),
	nil, 0
};

*/


// -----------------------
//  Plugin Initialisation 
// -----------------------

void PluginEntry (void)
{
	// register class definitions
	#if classExtension
		//REALRegisterClassExtension (&ExampleClass);
	#else
		//REALRegisterClass (&ExampleClass);
	#endif
	
	// register global method definitions
	for (int i = 0; i < sizeof(pluginMethodArray) / sizeof(REALmethodDefinition); ++i) {
		REALRegisterMethod (&pluginMethodArray[i]);
	}
}
